#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "vector.h"

void S_to_C(SPHERICAL s, CARTESIAN *c)
{
    c->x = s.r*sin(s.theta)*cos(s.phi);
    c->y = s.r*sin(s.theta)*sin(s.phi);
    c->z = s.r*cos(s.theta);
}

void C_to_S(CARTESIAN c, SPHERICAL *s)
{
    double r2;

    r2 = c.x*c.x + c.y*c.y + c.z*c.z;

    if (r2 == 0.0) {
        s->r = 0.0;
        s->theta = 0.0;
        s->phi = 0.0;
        return;
    }

    s->r = sqrt(r2);
    s->theta = acos(c.z / s->r);
    s->phi = atan2(c.y, c.x);
}

void P_to_S(POLAR p, SPHERICAL *s)
{
    s->phi = p.phi;
    s->r = sqrt(p.r*p.r + p.z*p.z);
    if (s->r > 0.0)
        s->theta = acos(p.z/s->r);
    else
        s->theta = 0.0;
}

void S_to_P(SPHERICAL s, POLAR *p)
{
    p->phi = s.phi;
    p->z = s.r * cos(s.theta);
    p->r = sqrt(s.r*s.r - p->z*p->z);
}

void P_to_C(POLAR p, CARTESIAN *c)
{
    c->x = p.r*cos(p.phi);
    c->y = p.r*sin(p.phi);
    c->z = p.z;
}

void C_to_P(CARTESIAN c, POLAR *p)
{
    p->r = sqrt(c.x*c.x + c.y*c.y);
    if (p->r == 0.0) {
        p->phi = 0.0;
    } else {
        p->phi = atan2(c.y, c.x);
    }
    p->z = c.z;
}

/* Here starts Cartesian functions */

void C_neg(CARTESIAN c, CARTESIAN *a)
{
    a->x = -c.x;
    a->y = -c.y;
    a->z = -c.z;
}

void C_line(CARTESIAN o, CARTESIAN v, double t, CARTESIAN *c)
{
    c->x = o.x + t*v.x;
    c->y = o.y + t*v.y;
    c->z = o.z + t*v.z;
}

void C_add(CARTESIAN c1, CARTESIAN c2, CARTESIAN *c)
{
    c->x = c1.x + c2.x;
    c->y = c1.y + c2.y;
    c->z = c1.z + c2.z;
}

void C_sub(CARTESIAN c1, CARTESIAN c2, CARTESIAN *c)
{
    c->x = c1.x - c2.x;
    c->y = c1.y - c2.y;
    c->z = c1.z - c2.z;
}

double C_dot(CARTESIAN c1, CARTESIAN c2)
{
    return (c1.x*c2.x + c1.y*c2.y + c1.z*c2.z);
}

void C_cross(CARTESIAN a, CARTESIAN b, CARTESIAN *c)
{
    c->x = a.y*b.z - a.z*b.y;
    c->y = a.z*b.x - a.x*b.z;
    c->z = a.x*b.y - a.y*b.x;
}

double C_magn(CARTESIAN c)
{
    return sqrt(c.x*c.x + c.y*c.y + c.z*c.z);
}

void C_norm(CARTESIAN c, CARTESIAN *a)
{
    double r;

    *a = c;
    r = sqrt(c.x*c.x + c.y*c.y + c.z*c.z);

    if (r != 0.0) {
        a->x = c.x/r;
        a->y = c.y/r;
        a->z = c.z/r;
    }
}

/* Here starts spherical functions */

void S_neg(SPHERICAL s, SPHERICAL *a)
{
    CARTESIAN c1, c2;

    S_to_C(s, &c1);
    C_neg(c1, &c2);
    C_to_S(c2, a);
}

void S_add(SPHERICAL s1, SPHERICAL s2, SPHERICAL *s)
{
    CARTESIAN c1, c2, c3;

    S_to_C(s1, &c1);
    S_to_C(s2, &c2);

    C_add(c1, c2, &c3);

    C_to_S(c3, s);
}

void S_sub(SPHERICAL s1, SPHERICAL s2, SPHERICAL *s)
{
    CARTESIAN c1, c2, c3;

    S_to_C(s1, &c1);
    S_to_C(s2, &c2);

    C_sub(c1, c2, &c3);

    C_to_S(c3, s);
}

double S_dot(SPHERICAL s1, SPHERICAL s2)
{
    double d;
    CARTESIAN c1, c2;

    S_to_C(s1, &c1);
    S_to_C(s2, &c2);

    d = C_dot(c1, c2);
  
    return d;
}

void S_cross(SPHERICAL s1, SPHERICAL s2, SPHERICAL *s)
{
    CARTESIAN c1, c2, c3;

    S_to_C(s1, &c1);
    S_to_C(s2, &c2);

    C_cross(c1, c2, &c3);

    C_to_S(c3, s);
}

double S_magn(SPHERICAL s)
{
    return s.r;
}

void S_norm(SPHERICAL s, SPHERICAL *a)
{
    *a = s;
    a->r = 1.0;
}

/* Here starts cylindrical functions */

void P_neg(POLAR p, POLAR *a)
{
    CARTESIAN c1, c2;

    P_to_C(p, &c1);
    C_neg(c1, &c2);
    C_to_P(c2, a);
}

void P_add(POLAR p1, POLAR p2, POLAR *p)
{
    CARTESIAN c1, c2, c3;

    P_to_C(p1, &c1);
    P_to_C(p2, &c2);

    C_add(c1, c2, &c3);

    C_to_P(c3, p);
}

void P_sub(POLAR p1, POLAR p2, POLAR *p)
{
    CARTESIAN c1, c2, c3;

    P_to_C(p1, &c1);
    P_to_C(p2, &c2);

    C_sub(c1, c2, &c3);

    C_to_P(c3, p);
}

double P_dot(POLAR p1, POLAR p2)
{
    double d;
    CARTESIAN c1, c2;

    P_to_C(p1, &c1);
    P_to_C(p2, &c2);

    d = C_dot(c1, c2);

    return d;
}

void P_cross(POLAR p1, POLAR p2, POLAR *p)
{
    CARTESIAN c1, c2, c3;

    P_to_C(p1, &c1);
    P_to_C(p2, &c2);

    C_cross(c1, c2, &c3);

    C_to_P(c3, p);
}

double P_magn(POLAR p)
{
    double d;

    d = sqrt(p.r*p.r + p.z*p.z);

    return d;
}

void P_norm(POLAR p, POLAR *a)
{
    double d;

    *a = p;
    d = sqrt(p.r*p.r + p.z*p.z);

    if (d != 0.0) {
        a->r /= d;
        a->z /= d;
    }
}
