#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include "config.h"
#ifndef WORDS_BIGENDIAN
#define BYTESWAP
#else
#undef BYTESWAP
#endif

#ifndef PRF_EXTERNAL
#define PRF_EXTERNAL
#endif
#include "prf.h"

#include "fits.h"

#define BITPIX		16
#define C			2.99792458e8

typedef short int WORD;
static char record[PHYSRECLEN];

static int AllKeys, NKeys;
static FITSptr fits;

int AllocFITS()
{
    if (fits == NULL) {
        fits = (FITSptr)malloc((NKeys+1)*sizeof(FITS));
    } else {
        fits = (FITSptr)realloc(fits, (NKeys+1)*sizeof(FITS));
    }
    if (fits == NULL) {
        fprintf(stderr, "Error: Memory allocation failure.\n");
        return 1;
    }
    return 0;
}

void FreeFITS()
{
   if (fits) free((FITSptr)fits);
   fits = NULL;
}

int save_FITS(const char *fname, int ndata, int size, void *data)
{
  int k, n, key, type, i;
  char keyword[KEYWORDLEN+1], argument[VALUELEN+1], comment[COMMENTLEN+1];
  char logrec[LOGRECLEN+1];
  char tmp[81];
  FILE *fp;
  int ncards=0;
  int bytes, len, left;
  char *ptr;
  
  void Info();
  int get_buffer_key();
  
  fp = fopen(fname, "w");
  if (fp == NULL) {
    fprintf(stderr, "Cannot open file '%s' for writing.\n", fname);
    return 1;
  }
  
  for (k=0; k<AllKeys; k++) {
    key = known[k].key;
    n = get_buffer_key(key);
    if (n < 0) continue;

    type = fits[n].type;
    
    switch (type) {
      case NVARTYPE:
        strcpy(argument, "                    ");
        break;
      case BOOLTYPE:
        if (fits[n].val.l == TRUE) {
          strcpy(argument, "                   T");
        } else {
          strcpy(argument, "                   F");
        }
        break;
      case CHARTYPE:
        sprintf(tmp, "'%s'", fits[n].val.str);
        for (i=0; i<VALUELEN; i++) {
          if (i < strlen(tmp))
            argument[i] = tmp[i];
          else
            argument[i] = ' ';
        }
        argument[VALUELEN] = '\0';
        break;
      case LONGTYPE:
        sprintf(argument, "%20ld", fits[n].val.l);
        break;
      case REALTYPE:
        sprintf(argument, "%20le", fits[n].val.d);
        break;
      case CPLXTYPE:
        sprintf(argument, "%10le%10le", fits[n].val.c[0], fits[n].val.c[1]);
        break;
      default:
        fprintf(stderr, "Invalid FITS argument type (%d)\n", type);
        continue;
        break;
    }
    
    for (i=0; i<KEYWORDLEN; i++) {
      if (i < strlen(known[k].text))
        keyword[i] = known[k].text[i];
      else
        keyword[i] = ' ';
    }
    keyword[KEYWORDLEN] = '\0';
    
    sprintf(tmp, "%s", known[k].comment);
    for (i=0; i<COMMENTLEN; i++) {
      if (i < strlen(tmp))
        comment[i] = tmp[i];
      else
        comment[i] = ' ';
    }
    comment[COMMENTLEN] = '\0';
#ifdef DEBUG
    printf("Writing header entry %2d (%s) of %d.\n", k+1, keyword, AllKeys);
#endif
    sprintf(logrec, "%s= %s / %s\n", keyword, argument, comment);
    if (ncards == 0) memset(record, '\0', PHYSRECLEN);
    memcpy(&record[ncards*LOGRECLEN], logrec, LOGRECLEN);
    ncards++;
    if (ncards == MAXCARDS) {	/* Physical record is full, write it! */
      len = fwrite(record, sizeof(char), PHYSRECLEN, fp);
      if (len != PHYSRECLEN) {
        fprintf(stderr, "Error when writing header record (%d)\n", len);
        fclose(fp);
        return (-1);
      }
      ncards = 0;
    }
  }
  
  if (ncards > 0) {		/* Write pending header record, if it exists */
    len = fwrite(record, sizeof(char), PHYSRECLEN, fp);
    if (len != PHYSRECLEN) {
      fprintf(stderr, "Error when writing pending header record (%d)\n", len);
      fclose(fp);
      return (-1);
    }
  }
  
  FreeFITS();
  
#ifdef DEBUG
  printf("Header written to file '%s'.\n", fname);
#endif

  ptr = (char *)data;
  bytes = size*ndata;
  i = 0;
  while (i < bytes) {
    memset(record, '\0', PHYSRECLEN);
    left = bytes - i;
    if (left > PHYSRECLEN) left = PHYSRECLEN;
    memcpy(record, ptr, left);
    ptr += left;
    i   += left;
#ifdef DEBUG
    printf("Writing data record (%d of %d)\n", i, bytes);
#endif
    len = fwrite(record, sizeof(char), PHYSRECLEN, fp);
    if (len != PHYSRECLEN) {
      fprintf(stderr, "Error when writing data record (%d)\n", len);
      fclose(fp);
      return (-1);
    }
#ifdef DEBUG
    printf("Data record of size %d written.\n", len);
#endif
  }
  
  sprintf(tmp, "Scan saved as '%s'.\n", fname);
  Info(tmp);
  
  fclose(fp);
  return 0;
}

int get_buffer_key(int key)
{
  int k;
  
  for (k=0; k<NKeys; k++)
    if (fits[k].key == key) return k;

  return (-1);
}

int add_FITS_header(int key, DATA data)
{
  int k, found=-1;
  
  if (AllocFITS()) {
    fprintf(stderr, "Error: NKeys= %d\n", NKeys);
    return -1;
  }
  
  AllKeys = sizeof(known)/sizeof(known[0]);
  for (k=0; k<AllKeys; k++) {
    if (key == known[k].key) {
      found = k;
      break;
    }
  }
  if (found < 0) {
    fprintf(stderr, "Warning: Unknown FITS keyword: %d\n", key);
    return 1;
  }
  
  fits[NKeys].key = key;
  fits[NKeys].type = known[found].type;
  fits[NKeys].val = data;
  NKeys++;
  return 1;
}

void addNvarCard(int key)
{
  DATA d;
  
  if (known[key].type != NVARTYPE) {
      fprintf(stderr, "Error! FITS key '%s' (no. %d) is not of NVARTYPE!\n",
              known[key].text, key);
      return;
  }
  
  strcpy(d.str, "               ");
  add_FITS_header(key, d);
}

void addBoolCard(int key, long value)
{
  DATA d;
  
  if (known[key].type != BOOLTYPE) {
      fprintf(stderr, "Error! FITS key '%s' (no. %d) is not of BOOLTYPE!\n",
              known[key].text, key);
      return;
  }
  
  d.l = value;
  add_FITS_header(key, d);
}

void addCharCard(int key, const char *value)
{
  DATA d;
  
  if (known[key].type != CHARTYPE) {
      fprintf(stderr, "Error! FITS key '%s' (no. %d) is not of CHARTYPE!\n",
              known[key].text, key);
      return;
  }
  
  strncpy(d.str, value, 18);
  d.str[18] = '\0';
 
  add_FITS_header(key, d);
}

void addLongCard(int key, long value)
{
  DATA d;
  
  if (known[key].type != LONGTYPE) {
      fprintf(stderr, "Error! FITS key '%s' (no. %d) is not of LONGTYPE!\n",
              known[key].text, key);
      return;
  }
  
  d.l = value;
  add_FITS_header(key, d);
}

void addRealCard(int key, double value)
{
  DATA d;
  
  if (known[key].type != REALTYPE) {
      fprintf(stderr, "Error! FITS key '%s' (no. %d) is not of REALTYPE!\n",
              known[key].text, key);
      return;
  }
  
  d.d = value;
  add_FITS_header(key, d);
}

void addCplxCard(int key, double value[2])
{
  DATA d;
  
  if (known[key].type != CPLXTYPE) {
      fprintf(stderr, "Error! FITS key '%s' (no. %d) is not of CPLXTYPE!\n",
              known[key].text, key);
      return;
  }
  
  d.c[0] = value[0];
  d.c[1] = value[1];
  add_FITS_header(key, d);
}

int Prf_to_Fits(int scan, int k, const char *fname,
                const char *source, const double *d)
{
  int n, ierr, size, nt = pos[k].trans;
  double max, min, bs, bz, offset;
  time_t clock;
  struct tm *date;
  char str[10], line[199];
  static WORD data[MAX_VEL];
  void initFits();
#ifdef BYTESWAP
  void swapbytes();
#endif
  initFits();
  
  addBoolCard(KW_SIMPLE,   TRUE);
  addLongCard(KW_BITPIX,   BITPIX);
  addLongCard(KW_NAXIS,    3);
  addLongCard(KW_NAXIS1,   (long int)prf.Nvel);
  addLongCard(KW_NAXIS2,   1);
  addLongCard(KW_NAXIS3,   1);
  addCharCard(KW_OBJECT,   source);
  addCharCard(KW_ORIGIN,   "mcp/prf");
  addCharCard(KW_OBSERVER, "mcp");
  sprintf(line, "%s(%d->%d)", prf.Molecule, tra[nt].u, tra[nt].l);
  line[18] = '\0';
  addCharCard(KW_LINE,     line);
  addCharCard(KW_INSTRUME, "");
  addCharCard(KW_TELESCOP, "");
  addRealCard(KW_RESTFREQ, prf.RestF*1.e9);
  addLongCard(KW_SCAN,     (long int)(1000 + scan));
  addRealCard(KW_OBSTIME,  1.0);
  addRealCard(KW_TSYS,     100.0);
  time(&clock);
  date = localtime(&clock);
  if (date) {
      sprintf(str, "%2d/%02d/%02d",
              date->tm_mday, date->tm_mon + 1, date->tm_year);
      str[9] = '\0';
      addCharCard(KW_DATE_OBS, str);
      addCharCard(KW_DATE, str);
      sprintf(str, "%2d:%02d:%02d", date->tm_hour, date->tm_min, date->tm_sec);
      str[9] = '\0';
      addCharCard(KW_TIME, str);
      addCharCard(KW_UTC,  str);
  }
  max = min = d[0];
  for (n=1; n<prf.Nvel; n++) {
    if (d[n] > max) max = d[n];
    if (d[n] < min) min = d[n];
  }
  bz = (max + min)/2.0;
  bs = (max - bz)/(pow(2.0, (double)(BITPIX-1)) - 1.0);
  addRealCard(KW_BSCALE,  bs);
  addRealCard(KW_BZERO,   bz);
  addCharCard(KW_BUNIT,   "K");
  addRealCard(KW_DATAMIN, min);
  addRealCard(KW_DATAMAX, max);
  addRealCard(KW_VLSR,    prf.Vlsr * 1.e3);
  addRealCard(KW_DELTAV,  prf.DeltaV * 1.e3);
  addCharCard(KW_CTYPE1,  "FREQ");
  addRealCard(KW_CRVAL1,  0.0);
  addRealCard(KW_CRPIX1,  prf.RefChan);
  addRealCard(KW_CDELT1,  prf.DeltaF*1.e9);
  addRealCard(KW_CROTA1,  0.0);
  addCharCard(KW_CTYPE2,  "RA");
  addRealCard(KW_CRVAL2,  0.0*RADTODEG);
  offset = pos[k].x * gSph.Rmax / pow(10.0, LOG_PC_CNST) / prf.Distance;
  addRealCard(KW_CDELT2,  offset*RADTODEG);
  addRealCard(KW_CRPIX2,  0.0*RADTODEG);
  addRealCard(KW_CROTA2,  0.0);
  addCharCard(KW_CTYPE3,  "DEC");
  addRealCard(KW_CRVAL3,  0.0*RADTODEG);
  offset = pos[k].y * gSph.Rmax / pow(10.0, LOG_PC_CNST) / prf.Distance;
  addRealCard(KW_CDELT3,  offset*RADTODEG);
  addRealCard(KW_CRPIX3,  0.0*RADTODEG);
  addRealCard(KW_CROTA3,  0.0);
  addNvarCard(KW_END);
  
  for (n=0; n<prf.Nvel; n++) {
    if (bs != 0.0)
      data[n] = (WORD)((d[n] - bz)/bs);
    else
      data[n] = 0;
#ifdef BYTESWAP
    /* The FITS standard expects the binary words with the MSB first */
    /* We have to swap them on e.g. an Intel CPU */
    swapbytes((char *)&data[n], sizeof(short int));
#endif
#ifdef DEEPDEBUG
    printf("Data[%3d] = %6d  (%12.5e)\n", n, data[n], (d[n]-bz)/bs);
#endif
  }
  
  n = (int)prf.Nvel;
  size = (int)sizeof(WORD);
  ierr = save_FITS(fname, n, size, (void *)data);
  
  return 0;
}

int assign_keyword(f)
FITS f;
{
  if (AllocFITS()) {
    fprintf(stderr, "Error: NKeys= %d\n", NKeys);
    return 1;
  }
  
  fits[NKeys] = f;
  NKeys++;
  return 0;
}

void *getKey(int key)
{
  int k;
  
  for (k=0; k<NKeys; k++) {
    if (key == fits[k].key) {
      switch (fits[k].type) {
        case NVARTYPE:
          return NULL;
        case BOOLTYPE:
          return &fits[k].val.l;
        case CHARTYPE:
          return fits[k].val.str;
        case LONGTYPE:
          return &fits[k].val.l;
        case REALTYPE:
          return &fits[k].val.d;
        case CPLXTYPE:
          return fits[k].val.c;
        default:
          return NULL;
          break;
      }
    }
  }
  
  return NULL;
}

/*
int read_FITS(fname, size, data)
char *fname;
int size;
void *data;
{
  int k, n, len, keycode;
  long ln, total;
  char keyword[KEYWORDLEN+1], argument[VALUELEN+1], comment[COMMENTLEN+1];
  char logrec[LOGRECLEN+1];
  FILE *fp;
  long *lptr;
  char *sptr;
  
  fp = fopen(fname, "r");
  if (fp == NULL) {
    fprintf(stderr, "Cannot open file '%s' for reading.\n", fname);
    return 1;
  }
  
  keycode = -1;
  while (keycode != KW_END) {
    len = fread(record, sizeof(char), PHYSRECLEN, fp);
    if (len != PHYSRECLEN) {
      fprintf(stderr, "Error reading header record [len=%d,%d]\n",
              len, PHYSRECLEN);
      fclose(fp);
      return 1;
    }
    for (n=0; n<MAXCARDS; n++) {
      strncpy(logrec, &record[n*LOGRECLEN], LOGRECLEN);
      logrec[LOGRECLEN] = '\0';
      sscanf(logrec, "%8s= %20s \ %s", keyword, argument, comment);
      memcpy(argument, &logrec[KEYWORDLEN + 2], VALUELEN);
      argument[VALUELEN] = '\0';
      memcpy(comment, &logrec[KEYWORDLEN + 2 + VALUELEN + 3], COMMENTLEN);
      comment[COMMENTLEN] = '\0';
      
      keycode = checkFitsKeyword(keyword, argument);
      if (keycode == KW_END) break;
    }
  }
  
  if ((lptr = getKey(KW_SIMPLE)) == NULL || *lptr != (long)TRUE) {
    fprintf(stderr, "I can only handle true SIMPLE.\n");
    fclose(fp);
    FreeFITS();
    return 1;
  }
  
  if ((lptr = getKey(KW_BITPIX)) == NULL || *lptr != (long)BITPIX) {
    fprintf(stderr, "I can only handle BITPIX=%d.\n", BITPIX);
    fclose(fp);
    FreeFITS();
    return 1;
  }
  
  if ((lptr = getKey(KW_NAXIS1)) == NULL) {
    fprintf(stderr, "Cannot find keyword 'NAXIS1'.\n");
    fclose(fp);
    FreeFITS();
    return 1;
  }
  
  if (*lptr > SCAN_CHANNELS) {
    fprintf(stderr, "Warning! Scan contains too many data channels: %ld\n", *lptr);
    *lptr = SCAN_CHANNELS;
  }
  
  total = (*lptr) * (long)size;
  sptr = (char *)data;
  ln = 0;  
  while (ln < total) {
    len = fread(record, sizeof(char), PHYSRECLEN, fp);
    ln += (long)len;
    if (len != PHYSRECLEN) {
      fprintf(stderr, "Error reading data record [%d,%d]\n", len, PHYSRECLEN);
      fclose(fp);
      FreeFITS();
      return 1;
    }
    if (ln < total)
      k = PHYSRECLEN;
    else
      k = PHYSRECLEN + (int)(total - ln);
    memcpy(sptr, record, k);
    sptr += k;
  }
  
  fclose(fp);
  FreeFITS();
  return 0;
}

int Fits_to_Scan(s, fname)
ScanPtr s;
char *fname;
{
  int n, size, ierr;
  static WORD data[SCAN_CHANNELS];
  double bs, bz, min, max, new_bz, new_bs;
  long *lptr;
  double *dptr;
  char *sptr, tmp[81];
  
  size = (int)sizeof(WORD);
  
  initFits();
  
  ierr = read_FITS(fname, size, data);
  if (ierr != 0) return 0;
  
  if ((lptr = getKey(KW_NAXIS1)) == NULL || *lptr > (long)SCAN_CHANNELS) {
    fprintf(stderr, "Error: Too many channels?\n");
    return 0;
  } else
    s->NChannel = (int)*lptr;
  if ((sptr = getKey(KW_OBJECT)) != NULL) strcpy(s->Object, sptr);
  if ((sptr = getKey(KW_OBSERVER)) != NULL) strcpy(s->Observer, sptr);
  if ((sptr = getKey(KW_LINE)) != NULL) strcpy(s->Line, sptr);
  if ((dptr = getKey(KW_RESTFREQ)) != NULL) s->RestFreq = *dptr/1.e6;
  if ((dptr = getKey(KW_SCAN)) != NULL) {
    s->ScanNo = (long)(*dptr);
  } else
    s->ScanNo = 1000;
  if ((sptr = getKey(KW_DATE_OBS)) != NULL) strcpy(s->DateObs, sptr);
  if ((dptr = getKey(KW_OBSTIME)) != NULL) s->ObsTime = *dptr;
  if ((sptr = getKey(KW_UTC)) != NULL) strcpy(s->UTC, sptr);
  if ((dptr = getKey(KW_BSCALE)) != NULL) bs = *dptr;
  if ((dptr = getKey(KW_BZERO)) != NULL) bz = *dptr;
  if ((dptr = getKey(KW_DATAMIN)) != NULL) min = *dptr;
  if ((dptr = getKey(KW_DATAMAX)) != NULL) max = *dptr;
  if ((dptr = getKey(KW_VLSR)) != NULL) s->VLsr = *dptr/1.e3;
  if ((dptr = getKey(KW_VELO_LSR)) != NULL) s->VLsr = *dptr/1.e3;
  if ((dptr = getKey(KW_DELTAV)) != NULL) s->DeltaV = *dptr/1.e3;
  
  if ((sptr = getKey(KW_CTYPE1)) == NULL || strncmp(sptr, "FREQ", 4) != 0) {
    fprintf(stderr, "Error: Keyword CTYPE1 must be 'FREQ' not '%s'.\n", sptr);
    return 0;
  }
  if ((dptr = getKey(KW_CDELT1)) != NULL) s->DeltaF = *dptr/1.e6;
  if ((dptr = getKey(KW_CRPIX1)) != NULL) s->RefChannel = (int)*dptr;
  
  if ((sptr = getKey(KW_CTYPE2)) == NULL || strncmp(sptr, "RA", 2) != 0) {
    fprintf(stderr, "Error: Keyword CTYPE2 must be 'RA' not '%s'.\n", sptr);
    return 0;
  }
  if ((dptr = getKey(KW_CRVAL2)) != NULL) s->Ra = *dptr/RADTODEG;
  if ((dptr = getKey(KW_CRPIX2)) != NULL) s->RaOff = *dptr/RADTODEG;
  if ((dptr = getKey(KW_CDELT2)) != NULL) s->RaSpacing = *dptr/RADTODEG;
  
  if ((sptr = getKey(KW_CTYPE3)) == NULL || strncmp(sptr, "DEC", 3) != 0) {
    fprintf(stderr, "Error: Keyword CTYPE3 must be 'DEC' not '%s'.\n", sptr);
    return 0;
  }
  if ((dptr = getKey(KW_CRVAL3)) != NULL) s->Dec = *dptr/RADTODEG;
  if ((dptr = getKey(KW_CRPIX3)) != NULL) s->DecOff = *dptr/RADTODEG;
  if ((dptr = getKey(KW_CDELT3)) != NULL) s->DecSpacing = *dptr/RADTODEG;
  
  new_bz = (max + min)/2.0;
  new_bs = (max - min)/2.0/(pow(2.0, (double)(BITPIX-1)) - 1.0);

  for (n=0; n<s->NChannel; n++) s->Data[n] = new_bz + (double)data[n] * new_bs;
  
  sprintf(tmp, "Scan #%ld (%d channels), read from '%s'.\n",
          s->ScanNo, s->NChannel, fname);
  Info(tmp);
  
  return 0;
}
*/

int checkFitsKeyword(char *keyword, char *argument)
{
  int k;
  FITS f;
  char *token;

  f.key = -1;
  for (k=0; k<AllKeys; k++) {
    if (strcmp(keyword, known[k].text) == 0) {
      f.key = known[k].key;
      f.type = known[k].type;
      switch (known[k].type) {
        case NVARTYPE:
          f.val.l = 0;
          break;
        case BOOLTYPE:
          if (strcmp(argument, "                   T") == 0) {
            f.val.l = TRUE;
          } else {
            f.val.l = FALSE;
          }
          break;
        case CHARTYPE:
          token = strtok(argument, "'");
          if (token != NULL)
            strcpy(f.val.str, token);
          else
            strcpy(f.val.str, argument);
          break;
        case LONGTYPE:
          sscanf(argument, "%ld", &f.val.l);
          break;
        case REALTYPE:
          sscanf(argument, "%lf", &f.val.d);
          break;
        case CPLXTYPE:
          sscanf(argument, "%10lf%10lf", &f.val.c[0], &f.val.c[1]);
          break;
        default:
          fprintf(stderr, "Warning! Unknown argument type: %d\n", known[k].type);
          f.key = -1;
          break;
      }
      break;
    }
  }
  
  if (f.key >= 0) {
    if (assign_keyword(f)) return -2;
  } else {
#ifdef DEBUG
    fprintf(stderr, "Warning! Unknown keyword: '%s'\n", keyword);
#endif
  }
  
  return f.key;
}

void initFits()
{
  AllKeys = sizeof(known)/sizeof(known[0]);
  NKeys = 0;
}

void Info(char *str)
{
  printf("%s", str);
}

#ifdef BYTESWAP
/* the following routine is needed for Intel type CPUs in order
to have the proper byte gender required by the FITS standard */

void swapbytes(char *p, int n)
{
    char swap;
    int i;

    for (i = 0; i < n/2; i++) {
	swap = p[i];
	p[i] = p[n-1-i];
	p[n-1-i] = swap;
    }
}
#endif
